// Generated by CoffeeScript 1.8.0
var Container, ModuleNotFound, construct, getArguments, runFactory, throwNotFound;

ModuleNotFound = require('./ModuleNotFound');

getArguments = require('./get-arguments');

construct = function(c, args) {
  var F;
  F = (function() {
    function F() {
      c.apply(this, args);
    }

    return F;

  })();
  F.prototype = c.prototype;
  return new F();
};

runFactory = function(factory, args) {
  return factory.apply(null, args);
};

throwNotFound = function(name, parent) {
  throw new ModuleNotFound(name, parent);
};

module.exports = Container = (function() {
  function Container(conf, _parent) {
    if (conf == null) {
      conf = {};
    }
    this._parent = _parent;
    this._modules = conf.modules || {};
    this._registrations = {};
    this._instances = {};
  }

  Container.prototype.factory = function(name, func) {
    if (!(func instanceof Function)) {
      throw new Error('A factory must be a function');
    }
    return this._register('factory', name, func);
  };

  Container.prototype.value = function(name, value) {
    return this._register('value', name, value);
  };

  Container.prototype["class"] = function(name, constructor) {
    if (!(constructor instanceof Function)) {
      throw new TypeError('A constructor must be a function');
    }
    return this._register('class', name, constructor);
  };

  Container.prototype.get = function(name) {
    if (!this.isRegistered(name)) {
      throwNotFound(name);
    }
    if (this._registeredAt(name) === 'local') {
      if (this._instances[name] == null) {
        this._instantiate(name);
      }
      return this._instances[name];
    } else {
      return this._parent.get(name);
    }
  };

  Container.prototype._registeredAt = function(name) {
    if (this._registrations[name] != null) {
      return 'local';
    } else if ((this._parent != null) && this._parent.isRegistered(name)) {
      return 'parent';
    }
  };

  Container.prototype.isRegistered = function(name) {
    return this._registeredAt(name) !== void 0;
  };

  Container.prototype.getArguments = function(name) {
    if (this._modules[name] != null) {
      return this._modules[name];
    } else {
      return getArguments(this._registrations[name].value);
    }
  };

  Container.prototype.loadAll = function() {
    if (this._parent != null) {
      this._parent.loadAll();
    }
    return Object.keys(this._registrations).forEach((function(_this) {
      return function(name) {
        if (_this._instances[name] == null) {
          return _this._instantiate(name);
        }
      };
    })(this));
  };

  Container.prototype._register = function(type, name, value) {
    return this._registrations[name] = {
      value: value,
      type: type
    };
  };

  Container.prototype._instantiate = function(name, parent) {
    var instance, module, type, value;
    module = this._registrations[name];
    if (module == null) {
      throwNotFound(name, parent);
    }
    type = module.type;
    value = module.value;
    instance = type === 'value' ? value : this._instantiateWithDependencies(name, value, type);
    this._instances[name] = instance;
    return instance;
  };

  Container.prototype._instantiateWithDependencies = function(name, value, type) {
    var args, dependencies;
    args = this.getArguments(name);
    dependencies = args.map((function(_this) {
      return function(depName) {
        if (_this._registeredAt(depName) === 'parent') {
          return _this._parent.get(depName);
        } else {
          return _this._instances[depName] || _this._instantiate(depName, name);
        }
      };
    })(this));
    if (type === 'factory') {
      return runFactory(value, dependencies);
    }
    if (type === 'class') {
      return construct(value, dependencies);
    }
  };

  return Container;

})();
